import sys
from typing import Callable, Type

from vedro.core import Dispatcher, Plugin, PluginConfig
from vedro.events import ArgParsedEvent, ArgParseEvent, CleanupEvent

__all__ = ("Terminator", "TerminatorPlugin",)


class TerminatorPlugin(Plugin):
    def __init__(self, config: Type["Terminator"], *, exit_fn: Callable[[int], None] = sys.exit):
        super().__init__(config)
        self._exit_fn = exit_fn
        self._no_scenarios_ok: bool = False

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(CleanupEvent, self.on_cleanup, priority=sys.maxsize)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        event.arg_parser.add_argument("--no-scenarios-ok",
                                      action="store_true",
                                      default=self._no_scenarios_ok,
                                      help="Exit with code 0 even if no scenarios are executed")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._no_scenarios_ok = event.args.no_scenarios_ok

    def on_cleanup(self, event: CleanupEvent) -> None:
        if event.report.interrupted:
            exc = event.report.interrupted.value
            if isinstance(exc, SystemExit) and (exc.code is not None):
                self._exit_fn(int(exc.code))
            else:
                self._exit_fn(130)
        elif event.report.failed > 0:
            self._exit_fn(1)
        elif not self._no_scenarios_ok and event.report.passed == 0:
            self._exit_fn(1)
        else:
            self._exit_fn(0)


class Terminator(PluginConfig):
    plugin = TerminatorPlugin
    description = "Handles exit status based on test results and interruptions"



from typing import Set, Type

from niltype import Nil

from vedro.core import Dispatcher, Plugin, PluginConfig
from vedro.core.exp.local_storage import LocalStorageFactory, create_local_storage
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ConfigLoadedEvent,
    ScenarioReportedEvent,
    StartupEvent,
)

__all__ = ("LastFailed", "LastFailedPlugin",)


class LastFailedPlugin(Plugin):
    def __init__(self, config: Type["LastFailed"], *,
                 local_storage_factory: LocalStorageFactory = create_local_storage) -> None:
        super().__init__(config)
        self._local_storage_factory = local_storage_factory
        self._last_failed = False
        self._failed_scenarios: Set[str] = set()

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ConfigLoadedEvent, self.on_config_loaded) \
                  .listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(StartupEvent, self.on_startup) \
                  .listen(ScenarioReportedEvent, self.on_scenario_reported) \
                  .listen(CleanupEvent, self.on_cleanup)

    def on_config_loaded(self, event: ConfigLoadedEvent) -> None:
        self._local_storage = self._local_storage_factory(self, event.config.project_dir)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        group = event.arg_parser.add_argument_group("Last Failed")
        group.add_argument("--last-failed", action="store_true", default=self._last_failed,
                           help="Run only last failed scenarios")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._last_failed = event.args.last_failed

    async def on_startup(self, event: StartupEvent) -> None:
        if not self._last_failed:
            return

        last_failed = await self._local_storage.get("last_failed")
        failed_scenarios = set(last_failed) if (last_failed is not Nil) else set()

        async for scenario in event.scheduler:
            if scenario.unique_id not in failed_scenarios:
                event.scheduler.ignore(scenario)

    def on_scenario_reported(self, event: ScenarioReportedEvent) -> None:
        if event.aggregated_result.is_failed():
            unique_id = event.aggregated_result.scenario.unique_id
            self._failed_scenarios.add(unique_id)

    async def on_cleanup(self, event: CleanupEvent) -> None:
        await self._local_storage.put("last_failed", list(self._failed_scenarios))
        await self._local_storage.flush()


class LastFailed(PluginConfig):
    plugin = LastFailedPlugin
    description = "Runs only the previously failed scenarios"



from hashlib import blake2b
from inspect import BoundArguments
from pathlib import Path
from typing import Any, List, Optional, Type, Union, cast

from .._scenario import Scenario
from ._virtual_step import VirtualStep

__all__ = ("VirtualScenario", "ScenarioInitError",)


class ScenarioInitError(Exception):
    """
    Raised when there is an error initializing a scenario.
    """
    pass


class VirtualScenario:
    """
    Represents a virtualized scenario with steps and metadata.

    This class wraps around an original Scenario class and its steps, providing
    additional metadata and functionality, such as unique identification, hashing,
    and skipping capabilities.
    """

    def __init__(self, orig_scenario: Type[Scenario], steps: List[VirtualStep], *,
                 project_dir: Path = Path.cwd()) -> None:
        """
        Initialize the VirtualScenario instance with the original scenario and steps.

        :param orig_scenario: The original scenario class.
        :param steps: A list of VirtualStep instances representing the steps in the scenario.
        :param project_dir: The project directory path.
        """
        self._orig_scenario = orig_scenario
        self._steps = steps
        # TODO: Make project_dir required in v2.0
        self._project_dir = project_dir.resolve()
        # TODO: Move path to constructor in v2.0
        self._path = Path(getattr(orig_scenario, "__file__", "."))
        self._is_skipped = False
        self._skip_reason: Union[str, None] = None

    @property
    def steps(self) -> List[VirtualStep]:
        """
        Get the list of steps in the scenario.

        :return: A list of VirtualStep instances.
        """
        return self._steps

    @property
    def unique_id(self) -> str:
        """
        Get a unique identifier for the scenario.

        :return: A string representing the unique ID of the scenario.
        """
        unique_id = f"{self.rel_path}::{self.name}"
        if self.template_index is not None:
            unique_id += f"#{self.template_index}"
        return unique_id

    @property
    def unique_hash(self) -> str:
        """
        Get a unique hash for the scenario.

        :return: A string representing the unique hash of the scenario.
        """
        return blake2b(self.unique_id.encode(), digest_size=20).hexdigest()

    @property
    def template_index(self) -> Union[int, None]:
        """
        Get the template index of the current scenario in the templated scenario.

        :return: An integer representing the template index, or None if not applicable.
        """
        idx = getattr(self._orig_scenario, "__vedro__template_index__", None)
        return cast(Union[int, None], idx)

    @property
    def template_total(self) -> Union[int, None]:
        """
        Get the total number of scenarios in the templated scenario.

        :return: An integer representing the total number of scenarios, or None if not applicable.
        """
        idx = getattr(self._orig_scenario, "__vedro__template_total__", None)
        return cast(Union[int, None], idx)

    @property
    def template_args(self) -> Union[BoundArguments, None]:
        """
        Get the bound arguments for the templated scenario.

        :return: A BoundArguments object representing the template arguments,
                 or None if not applicable.
        """
        args = getattr(self._orig_scenario, "__vedro__template_args__", None)
        return cast(Union[BoundArguments, None], args)

    @property
    def path(self) -> Path:
        """
        Get the path to the scenario file.

        :return: A Path object representing the path to the scenario file.
        """
        return self._path

    @property
    def rel_path(self) -> Path:
        """
        Get the relative path to the scenario file from the project directory.

        :return: A Path object representing the relative path to the scenario file.
        """
        return self._path.relative_to(self._project_dir)

    @property
    def name(self) -> str:
        """
        Get the name of the scenario.

        :return: A string representing the name of the scenario.
        """
        return getattr(self._orig_scenario, "__vedro__template_name__",
                       self._orig_scenario.__name__)

    @property
    def subject(self) -> str:
        """
        Get the subject of the scenario.

        :return: A string representing the subject of the scenario.
        :raises ValueError: If the subject cannot be formatted with the template arguments.
        """
        subject = getattr(self._orig_scenario, "subject", None)
        if not isinstance(subject, str) or subject.strip() == "":
            subject = self._path.stem.replace("_", " ")

        if not self.template_args:
            return subject

        try:
            return subject.format(**self.template_args.arguments)
        except Exception as exc:
            message = f'Can\'t format subject "{subject}" at "{self.rel_path}" ({exc})'
            raise ValueError(message) from None

    @property
    def namespace(self) -> str:
        """
        Get the namespace of the scenario.

        :return: A string representing the namespace of the scenario.
        """
        parts = self.rel_path.parts[1:-1]
        return str(Path(*parts)) if parts else ""

    def skip(self, reason: Optional[str] = None) -> None:
        """
        Mark the scenario as skipped with an optional reason.

        :param reason: The reason for skipping the scenario.
        """
        self._is_skipped = True
        if reason:
            self._skip_reason = reason

    @property
    def skip_reason(self) -> Optional[str]:
        """
        Get the reason for skipping the scenario.

        :return: A string representing the skip reason or None if not skipped.
        """
        return self._skip_reason

    def is_skipped(self) -> bool:
        """
        Check if the scenario is marked as skipped.

        :return: A boolean indicating if the scenario is skipped.
        """
        return self._is_skipped

    def __call__(self) -> Scenario:
        """
        Initialize and return an instance of the original scenario.

        :return: An instance of the original scenario.
        :raises ScenarioInitError: If the scenario cannot be initialized.
        """
        try:
            return self._orig_scenario()
        except Exception as exc:
            message = f'Can\'t initialize scenario "{self.subject}" at "{self.rel_path}" ({exc!r})'
            raise ScenarioInitError(message) from None

    def __repr__(self) -> str:
        """
        Return a string representation of the VirtualScenario instance.

        :return: A string representing the VirtualScenario instance.
        """
        return f"<{self.__class__.__name__} {str(self.rel_path)!r}>"

    def __eq__(self, other: Any) -> bool:
        """
        Check equality with another object.

        :param other: The other object to compare with.
        :return: A boolean indicating if the other object is equal to this instance.
        """
        return isinstance(other, self.__class__) and (self.__dict__ == other.__dict__)



from enum import Enum

__all__ = ("ScenarioStatus",)


class ScenarioStatus(Enum):
    """
    Enumeration of possible states for `ScenarioResult` to indicate the current status
    of a test scenario.

    For more information, refer to https://vedro.io/docs/core/scenario-status.
    """

    # Indicates the scenario is queued and waiting for execution.
    PENDING = "PENDING"

    # Signifies the scenario has completed successfully with all assertions validated.
    PASSED = "PASSED"

    # Marks the scenario as unsuccessful due to failed assertions or unexpected errors.
    FAILED = "FAILED"

    # Represents scenarios that are deliberately not executed.
    SKIPPED = "SKIPPED"



from pathlib import Path
from typing import List, Optional, Set

from vedro.core import ScenarioFinder, ScenarioLoader, ScenarioOrderer, VirtualScenario
from vedro.core.scenario_discoverer import ScenarioDiscoverer, create_vscenario

__all__ = ("SelectiveScenarioDiscoverer",)


class SelectiveScenarioDiscoverer(ScenarioDiscoverer):
    def __init__(self,
                 finder: ScenarioFinder, loader: ScenarioLoader, orderer: ScenarioOrderer, *,
                 selected_paths: Optional[Set[Path]] = None) -> None:
        super().__init__(finder, loader, orderer)
        self._selected_paths = selected_paths

    async def discover(self, root: Path) -> List[VirtualScenario]:
        scenarios = []
        async for path in self._finder.find(root):
            if not self._is_path_selected(path):
                continue
            loaded = await self._loader.load(path)
            for scn in loaded:
                scenarios.append(create_vscenario(scn))
        return await self._orderer.sort(scenarios)

    def _is_path_selected(self, path: Path) -> bool:
        if self._selected_paths is None or len(self._selected_paths) == 0:
            return True
        abs_path = path.absolute()
        return any(self._is_relative_to(abs_path, selected) for selected in self._selected_paths)

    def _is_relative_to(self, path: Path, parent: Path) -> bool:
        try:
            path.relative_to(parent)
        except ValueError:
            return False
        else:
            return True



import json
import platform
import time
import urllib.request
from threading import Thread
from typing import Any, Dict, Tuple, Type, Union

from niltype import Nil

import vedro
from vedro.core import Dispatcher, Plugin, PluginConfig
from vedro.core.exp.local_storage import LocalStorageFactory, create_local_storage
from vedro.events import CleanupEvent, ConfigLoadedEvent, StartupEvent

__all__ = ("SystemUpgrade", "SystemUpgradePlugin",)


class SystemUpgradePlugin(Plugin):
    def __init__(self, config: Type["SystemUpgrade"], *,
                 local_storage_factory: LocalStorageFactory = create_local_storage) -> None:
        super().__init__(config)
        self._api_url = config.api_url
        self._request_timeout = config.api_request_timeout
        self._local_storage_factory = local_storage_factory
        self._thread: Union[Thread, None] = None
        self._latest_version: Union[str, None] = None
        self._last_request_ts: Union[int, None] = None
        self._last_request_error: Union[str, None] = None
        self._cur_version = vedro.__version__
        self._check_interval = config.update_check_interval

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ConfigLoadedEvent, self.on_config_loaded) \
                  .listen(StartupEvent, self.on_startup) \
                  .listen(CleanupEvent, self.on_cleanup)

    def on_config_loaded(self, event: ConfigLoadedEvent) -> None:
        self._local_storage = self._local_storage_factory(self, event.config.project_dir)

    async def on_startup(self, event: StartupEvent) -> None:
        last_request = await self._local_storage.get("last_request_ts")
        if (last_request is not Nil) and (last_request + self._check_interval > self._now()):
            return

        self._thread = Thread(target=self._get_latest_version, daemon=True)
        self._thread.start()

    def _is_up_to_date(self, cur_version: str, new_version: str) -> bool:
        def parse_version(version: str) -> Tuple[str, ...]:
            return tuple(x.zfill(8) for x in version.split("."))
        return bool(parse_version(cur_version) >= parse_version(new_version))

    async def on_cleanup(self, event: CleanupEvent) -> None:
        if self._thread:
            self._thread.join(0.0)
            self._thread = None

        if self._last_request_ts is None:
            self._last_request_ts = self._now()
            self._last_request_error = f"Request aborted (timeout {self._request_timeout})"

        await self._local_storage.put("last_request_ts", self._last_request_ts)
        await self._local_storage.put("last_request_error", self._last_request_error)
        await self._local_storage.flush()

        if self._latest_version:
            if not self._is_up_to_date(self._cur_version, self._latest_version):
                event.report.add_summary(
                    f"(!) Vedro update available: {self._cur_version} → {self._latest_version}"
                )

    def _send_request(self, url: str, *, headers: Dict[str, str], timeout: float) -> Any:
        request = urllib.request.Request(url, headers=headers)
        with urllib.request.urlopen(request, timeout=timeout) as response:
            return json.loads(response.read())

    def _get_user_agent(self) -> str:
        python_version = platform.python_version()
        platform_info = platform.platform(terse=True)
        return f"Vedro/{self._cur_version} (Python/{python_version}; {platform_info})"

    def _get_latest_version(self) -> None:
        url = f"{self._api_url}/v1/latest-version"
        headers = {"User-Agent": self._get_user_agent()}
        try:
            response = self._send_request(url, headers=headers, timeout=self._request_timeout)
        except BaseException as e:
            self._last_request_error = str(e)
            self._last_request_ts = self._now()
            return
        if isinstance(response, dict) and ("version" in response):
            self._latest_version = response["version"]
            self._last_request_ts = self._now()

    def _now(self) -> int:
        return int(time.time())


class SystemUpgrade(PluginConfig):
    plugin = SystemUpgradePlugin
    description = "Checks for Vedro updates and notifies when a newer version is available"

    # API URL for checking the latest version
    api_url: str = "https://api.vedro.io"

    # Timeout for the API request in seconds
    api_request_timeout: float = 1.0

    # Interval at which the update check should be done, specified in seconds
    update_check_interval: int = 3600 * 16



from pathlib import Path
from typing import Type

import cabina

__all__ = ("Config", "Section", "ConfigType",)


class Section(cabina.Section):
    pass


class Config(cabina.Config, Section):
    # 'path' refers to the location of the configuration file.
    # By default, it's set to the location of this file itself (__file__).
    # In practice, this path is dynamically determined at runtime,
    # typically pointing to 'vedro.cfg.py' located in the project directory.
    path: Path = Path(__file__)

    # The `project_dir` property is designated as the root directory of the project,
    # serving as a reference point for relative paths and project-related file operations.
    # Although it defaults to the directory from which the application is executed,
    # this setting can be overridden via the `--project-dir` argument.
    project_dir: Path = Path.cwd()

    class Registry(Section):
        pass

    class Plugins(Section):
        pass


ConfigType = Type[Config]



from re import fullmatch
from typing import Set, Union, cast

from pyparsing import Literal
from pyparsing import ParserElement as Parser
from pyparsing import ParseResults, Regex, infixNotation, opAssoc
from pyparsing.exceptions import ParseException

from .._tag_matcher import TagMatcher
from ._logic_ops import And, Expr, Not, Operand, Or

__all__ = ("LogicTagMatcher",)


class LogicTagMatcher(TagMatcher):
    tag_pattern = r'(?!and$|or$|not$)[A-Za-z_][A-Za-z0-9_]*'

    def __init__(self, expr: str) -> None:
        super().__init__(expr)
        operand = Regex(self.tag_pattern).setParseAction(self._create_tag)
        self._parser = infixNotation(operand, [
            (Literal("not"), 1, opAssoc.RIGHT, self._create_not),
            (Literal("and"), 2, opAssoc.LEFT, self._create_and),
            (Literal("or"), 2, opAssoc.LEFT, self._create_or),
        ])
        self._grammar: Union[Expr, None] = None

    def match(self, tags: Set[str]) -> bool:
        if self._grammar is None:
            self._grammar = self._parse(self._parser, self._expr)
        return self._grammar(tags)

    def validate(self, tag: str) -> bool:
        if not isinstance(tag, str):
            raise TypeError(f"Tag must be a str, got {type(tag)}")

        res = fullmatch(self.tag_pattern, tag)
        if res is None:
            raise ValueError(f"Tag must match regex {self.tag_pattern!r}")

        return True

    def _create_tag(self, orig: str, location: int, tokens: ParseResults) -> Expr:
        tag = tokens[0]
        return Operand(tag)

    def _create_not(self, orig: str, location: int, tokens: ParseResults) -> Expr:
        operand = tokens[0][-1]
        return Not(operand)

    def _create_and(self, orig: str, location: int, tokens: ParseResults) -> Expr:
        left = tokens[0][0]
        right = tokens[0][-1]
        return And(left, right)

    def _create_or(self, orig: str, location: int, tokens: ParseResults) -> Expr:
        left = tokens[0][0]
        right = tokens[0][-1]
        return Or(left, right)

    def _parse(self, grammar: Parser, expr: str) -> Expr:
        try:
            results = grammar.parse_string(expr, parse_all=True)
        except ParseException as e:
            raise ValueError(f"Invalid tag expr {expr!r}. Error: {str(e)}") from None

        if len(results) == 0:
            raise ValueError(f"Invalid tag expr {expr!r}")
        return cast(Expr, results[0])



import asyncio
from typing import Any, Callable, Coroutine, Type, Union

from vedro.core import ConfigType, Dispatcher, Plugin, PluginConfig, ScenarioScheduler
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ConfigLoadedEvent,
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    ScenarioRunEvent,
    ScenarioSkippedEvent,
    StartupEvent,
)

from ._scheduler import RerunnerScenarioScheduler

__all__ = ("Rerunner", "RerunnerPlugin",)

SleepType = Callable[[float], Coroutine[Any, Any, None]]


class RerunnerPlugin(Plugin):
    def __init__(self, config: Type["Rerunner"], *, sleep: SleepType = asyncio.sleep) -> None:
        super().__init__(config)
        self._scheduler_factory = config.scheduler_factory
        self._sleep = sleep
        self._reruns: int = 0
        self._reruns_delay: float = 0.0
        self._global_config: Union[ConfigType, None] = None
        self._scheduler: Union[ScenarioScheduler, None] = None
        self._rerun_scenario_id: Union[str, None] = None
        self._reran: int = 0
        self._times: int = 0

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ConfigLoadedEvent, self.on_config_loaded) \
                  .listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(StartupEvent, self.on_startup) \
                  .listen(ScenarioRunEvent, self.on_scenario_execute) \
                  .listen(ScenarioSkippedEvent, self.on_scenario_execute) \
                  .listen(ScenarioPassedEvent, self.on_scenario_end) \
                  .listen(ScenarioFailedEvent, self.on_scenario_end) \
                  .listen(CleanupEvent, self.on_cleanup)

    def on_config_loaded(self, event: ConfigLoadedEvent) -> None:
        self._global_config = event.config

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        group = event.arg_parser.add_argument_group("Rerunner")
        help_message = ("Number of times to rerun failed scenarios (default: 0). "
                        "The resulting scenario status based on the majority status of reruns")
        group.add_argument("--reruns", type=int, default=self._reruns, help=help_message)
        group.add_argument("--reruns-delay", type=float, default=self._reruns_delay,
                           help="Delay in seconds between reruns (default: 0.0s)")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._reruns = event.args.reruns
        self._reruns_delay = event.args.reruns_delay

        if self._reruns < 0:
            raise ValueError("--reruns must be >= 0")

        if self._reruns_delay < 0.0:
            raise ValueError("--reruns-delay must be >= 0.0")

        if (self._reruns_delay > 0.0) and (self._reruns < 1):
            raise ValueError("--reruns-delay must be used with --reruns > 0")

        if self._is_rerunning_enabled():
            assert self._global_config is not None  # for type checking
            self._global_config.Registry.ScenarioScheduler.register(self._scheduler_factory, self)

    def on_startup(self, event: StartupEvent) -> None:
        self._scheduler = event.scheduler

    async def on_scenario_execute(self,
                                  event: Union[ScenarioRunEvent, ScenarioSkippedEvent]) -> None:
        if not self._is_rerunning_enabled():
            return

        scenario = event.scenario_result.scenario
        if (self._rerun_scenario_id == scenario.unique_id) and (self._reruns_delay > 0.0):
            await self._sleep(self._reruns_delay)

    async def on_scenario_end(self,
                              event: Union[ScenarioPassedEvent, ScenarioFailedEvent]) -> None:
        if not self._is_rerunning_enabled():
            return
        assert isinstance(self._scheduler, RerunnerScenarioScheduler)  # for type checking

        scenario = event.scenario_result.scenario
        if scenario.unique_id != self._rerun_scenario_id:
            self._rerun_scenario_id = scenario.unique_id

            if event.scenario_result.is_failed():
                self._reran += 1
                for _ in range(self._reruns):
                    self._scheduler.schedule(scenario)
                    self._times += 1

    def on_cleanup(self, event: CleanupEvent) -> None:
        if not self._is_rerunning_enabled():
            return
        message = self._get_summary_message()
        event.report.add_summary(message)

    def _is_rerunning_enabled(self) -> bool:
        return self._reruns > 0

    def _get_summary_message(self) -> str:
        ss = "" if self._reran == 1 else "s"
        ts = "" if self._times == 1 else "s"
        message = f"rerun {self._reran} scenario{ss}, {self._times} time{ts}"
        if self._reruns_delay:
            message += f", with delay {self._reruns_delay!r}s"
        return message


class Rerunner(PluginConfig):
    plugin = RerunnerPlugin
    description = "Reruns failed scenarios a specified number of times"

    # Scheduler that will be used to create aggregated result for rerun scenarios
    scheduler_factory: Type[ScenarioScheduler] = RerunnerScenarioScheduler



import importlib
import importlib.util
from importlib.abc import Loader
from importlib.machinery import ModuleSpec
from keyword import iskeyword
from pathlib import Path
from types import ModuleType
from typing import cast

from ._module_loader import ModuleLoader

__all__ = ("ModuleFileLoader",)


class ModuleFileLoader(ModuleLoader):
    """
    Loads a module from a file path with optional validation of module names.

    This class implements the ModuleLoader abstract base class, providing functionality
    to load a module from a specified file path. Optionally, it can validate the module
    names derived from the file paths to ensure they are valid Python identifiers.
    """

    def __init__(self, *, validate_module_names: bool = False) -> None:
        """
        Initialize the ModuleFileLoader with optional module name validation.

        :param validate_module_names: Flag to indicate whether module names should be
                                      validated. Defaults to False.
        """
        self._validate_module_names = validate_module_names

    async def load(self, path: Path) -> ModuleType:
        """
        Load a module from the specified file path.

        :param path: The file path to load the module from.
        :return: The loaded module.
        :raises ValueError: If the module name derived from the path is invalid and
                            validation is enabled.
        :raises ModuleNotFoundError: If the module spec could not be created from the path.
        """
        spec = self._spec_from_path(path)
        module = self._module_from_spec(spec)
        self._exec_module(cast(Loader, spec.loader), module)
        return module

    def _path_to_module_name(self, path: Path) -> str:
        """
        Convert a file path to a module name, optionally validating the name.

        :param path: The file path to convert.
        :return: The derived module name.
        :raises ValueError: If the module name derived from the path is invalid and
                            validation is enabled.
        """
        parts = path.with_suffix("").parts

        # Handle absolute paths by ignoring the root part (e.g., '/' or 'C:\')
        if path.is_absolute():
            parts = parts[1:]

        if self._validate_module_names:
            for part in parts:
                if not self._is_valid_identifier(part):
                    raise ValueError(
                        f"The module name derived from the path '{path}' is invalid "
                        f"due to the segment '{part}'. A valid module name should "
                        "start with a letter or underscore, contain only letters, "
                        "digits, or underscores, and not be a Python keyword."
                    )
        return ".".join(parts)

    def _is_valid_identifier(self, name: str) -> bool:
        """
        Check if a string is a valid Python identifier and not a keyword.

        :param name: The string to check.
        :return: True if the string is a valid identifier, False otherwise.
        """
        return name.isidentifier() and not iskeyword(name)

    def _spec_from_path(self, path: Path) -> ModuleSpec:
        """
        Create a module specification from a file path.

        :param path: The file path to create the spec from.
        :return: The created module specification.
        :raises ModuleNotFoundError: If the module spec could not be created from the path.
        """
        module_name = self._path_to_module_name(path)
        spec = importlib.util.spec_from_file_location(module_name, path)
        if spec is None:
            raise ModuleNotFoundError(module_name)
        return spec

    def _module_from_spec(self, spec: ModuleSpec) -> ModuleType:
        """
        Create a module from a module specification.

        :param spec: The module specification to create the module from.
        :return: The created module.
        """
        return importlib.util.module_from_spec(spec)

    def _exec_module(self, loader: Loader, module: ModuleType) -> None:
        """
        Execute a module using its loader.

        :param loader: The loader to use for executing the module.
        :param module: The module to execute.
        """
        loader.exec_module(module)



import ast
import inspect
from importlib.abc import Loader
from types import ModuleType
from typing import Any, cast, final

from vedro.core import ModuleFileLoader

from ._node_assert_rewriter import NodeAssertRewriter
from ._typing import override

__all__ = ("AssertRewriterLoader",)


@final
class AssertRewriterLoader(ModuleFileLoader):
    """
    Loads and rewrites Python modules to use custom assertion methods.

    This class extends ModuleFileLoader to rewrite assert statements in Python
    modules to use a specified assertion tool with custom assertion methods.
    """

    assert_module = "vedro.plugins.assert_rewriter"
    assert_tool = "assert_"
    assert_methods = {
        ast.Eq: "assert_equal",
        ast.NotEq: "assert_not_equal",

        ast.Lt: "assert_less",
        ast.LtE: "assert_less_equal",
        ast.Gt: "assert_greater",
        ast.GtE: "assert_greater_equal",

        ast.Is: "assert_is",
        ast.IsNot: "assert_is_not",

        ast.In: "assert_in",
        ast.NotIn: "assert_not_in",

        type(None): "assert_truthy",
    }

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """
        Initialize the AssertRewriterModuleLoader.

        :param args: Positional arguments for the base class.
        :param kwargs: Keyword arguments for the base class.
        """
        super().__init__(*args, **kwargs)
        self._assert_rewriter = NodeAssertRewriter(
            assert_tool=self.assert_tool,
            assert_methods=self.assert_methods
        )

    @override
    def _exec_module(self, loader: Loader, module: ModuleType) -> None:
        """
        Execute the module after rewriting its assert statements.

        :param loader: The loader used to load the module.
        :param module: The module to execute.
        """
        source_code = inspect.getsource(module)

        tree = ast.parse(source_code)
        rewritten_tree = self._rewrite_tree(tree)

        transformed = compile(rewritten_tree, module.__file__, "exec")  # type: ignore
        exec(transformed, module.__dict__)

    def _rewrite_tree(self, tree: ast.Module) -> ast.Module:
        """
        Rewrite the AST of the module to replace assert statements.

        :param tree: The AST of the module.
        :return: The rewritten AST of the module.
        """
        rewritten_tree = self._assert_rewriter.visit(tree)

        import_stmt = ast.ImportFrom(
            module=self.assert_module,
            names=[ast.alias(name=self.assert_tool, asname=None)],
            level=0
        )
        ast.fix_missing_locations(import_stmt)
        rewritten_tree.body.insert(0, import_stmt)

        return cast(ast.Module, rewritten_tree)



from abc import ABC
from pathlib import Path
from typing import Any

__all__ = ("Artifact", "MemoryArtifact", "FileArtifact",)


class Artifact(ABC):
    """
    The base class for representing artifacts in a system.

    An artifact in this context is a piece of data generated during the execution
    of a scenario or a step. It can be anything from log files, screenshots, to data dumps.
    This class serves as an abstract base class for different types of artifacts,
    such as MemoryArtifact and FileArtifact.
    """
    pass


class MemoryArtifact(Artifact):
    """
    Represents an artifact that is stored in memory.
    """

    def __init__(self, name: str,  mime_type: str, data: bytes) -> None:
        """
        Initialize a MemoryArtifact with a name, MIME type, and data.

        :param name: The name of the artifact.
        :param mime_type: The MIME type of the data.
        :param data: The actual data in bytes.
        """
        if not isinstance(data, bytes):
            raise TypeError("'data' must be of type bytes")
        self._name = name
        self._data = data
        self._mime_type = mime_type

    @property
    def name(self) -> str:
        """
        Get the name of the artifact.

        :return: The name of the artifact.
        """
        return self._name

    @property
    def mime_type(self) -> str:
        """
        Get the MIME type of the artifact data.

        :return: The MIME type as a string.
        """
        return self._mime_type

    @property
    def data(self) -> bytes:
        """
        Get the data stored in the artifact.

        :return: The data as bytes.
        """
        return self._data

    def __repr__(self) -> str:
        """
        Represent the MemoryArtifact as a string.

        :return: A string representation of the MemoryArtifact.
        """
        size = len(self._data)
        return f"{self.__class__.__name__}<{self._name!r}, {self._mime_type!r}, size={size}>"

    def __eq__(self, other: Any) -> bool:
        """
        Check equality with another MemoryArtifact.

        :param other: The other MemoryArtifact to compare with.
        :return: True if the other artifact is equal to this one, False otherwise.
        """
        return isinstance(other, self.__class__) and (self.__dict__ == other.__dict__)


class FileArtifact(Artifact):
    """
    Represents an artifact that is stored as a file on the filesystem.
    """

    def __init__(self, name: str, mime_type: str, path: Path) -> None:
        """
        Initialize a FileArtifact with a name, MIME type, and file path.

        :param name: The name of the artifact.
        :param mime_type: The MIME type of the file.
        :param path: The path to the file.
        """
        if not isinstance(path, Path):
            raise TypeError("'path' must be an instance of pathlib.Path")
        self._name = name
        self._path = path
        self._mime_type = mime_type

    @property
    def name(self) -> str:
        """
        Get the name of the artifact.

        :return: The name of the artifact.
        """
        return self._name

    @property
    def mime_type(self) -> str:
        """
        Get the MIME type of the artifact data.

        :return: The MIME type as a string.
        """
        return self._mime_type

    @property
    def path(self) -> Path:
        """
        Get the file path of the artifact.

        :return: The path of the file as a Path object.
        """
        return self._path

    def __repr__(self) -> str:
        """
        Represent the FileArtifact as a string.

        :return: A string representation of the FileArtifact.
        """
        return f"{self.__class__.__name__}<{self._name!r}, {self._mime_type!r}, {self._path!r}>"

    def __eq__(self, other: Any) -> bool:
        """
        Check equality with another FileArtifact.

        :param other: The other FileArtifact to compare with.
        :return: True if the other artifact is equal to this one, False otherwise.
        """
        return isinstance(other, self.__class__) and (self.__dict__ == other.__dict__)



import shutil
from collections import deque
from pathlib import Path
from typing import Deque, Type, Union

from vedro.core import (
    Artifact,
    ConfigType,
    Dispatcher,
    FileArtifact,
    MemoryArtifact,
    Plugin,
    PluginConfig,
    ScenarioResult,
    StepResult,
)
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    ConfigLoadedEvent,
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    ScenarioReportedEvent,
    ScenarioRunEvent,
    StepFailedEvent,
    StepPassedEvent,
)

__all__ = ("Artifacted", "ArtifactedPlugin",
           "attach_artifact", "attach_step_artifact", "attach_scenario_artifact",)


_scenario_artifacts: Deque[Artifact] = deque()
_step_artifacts: Deque[Artifact] = deque()


def attach_scenario_artifact(artifact: Artifact) -> None:
    _scenario_artifacts.append(artifact)


def attach_step_artifact(artifact: Artifact) -> None:
    _step_artifacts.append(artifact)


def attach_artifact(artifact: Artifact) -> None:
    attach_step_artifact(artifact)


class ArtifactedPlugin(Plugin):
    def __init__(self, config: Type["Artifacted"], *,
                 scenario_artifacts: Deque[Artifact] = _scenario_artifacts,
                 step_artifacts: Deque[Artifact] = _step_artifacts) -> None:
        super().__init__(config)
        self._scenario_artifacts = scenario_artifacts
        self._step_artifacts = step_artifacts
        self._save_artifacts = config.save_artifacts
        self._artifacts_dir = config.artifacts_dir
        self._add_artifact_details = config.add_artifact_details
        self._global_config: Union[ConfigType, None] = None

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ConfigLoadedEvent, self.on_config_loaded) \
                  .listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(ScenarioRunEvent, self.on_scenario_run) \
                  .listen(StepPassedEvent, self.on_step_end) \
                  .listen(StepFailedEvent, self.on_step_end) \
                  .listen(ScenarioPassedEvent, self.on_scenario_end) \
                  .listen(ScenarioFailedEvent, self.on_scenario_end) \
                  .listen(ScenarioReportedEvent, self.on_scenario_reported)

    def on_config_loaded(self, event: ConfigLoadedEvent) -> None:
        self._global_config = event.config

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        group = event.arg_parser.add_argument_group("Artifacted")

        group.add_argument("--save-artifacts", action="store_true",
                           default=self._save_artifacts,
                           help="Save artifacts to the file system")
        group.add_argument("--artifacts-dir", type=Path, default=None,
                           help=("Specify the directory path for saving artifacts "
                                 f"(default: '{self._artifacts_dir}')"))

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._save_artifacts = event.args.save_artifacts
        if not self._save_artifacts:
            if event.args.artifacts_dir is not None:
                raise ValueError(
                    "Artifacts directory cannot be specified when artifact saving is disabled")
            return
        self._artifacts_dir = event.args.artifacts_dir or self._artifacts_dir

        project_dir = self._get_project_dir()
        if not self._artifacts_dir.is_absolute():
            self._artifacts_dir = (project_dir / self._artifacts_dir).resolve()
        if not self._is_relative_to(self._artifacts_dir, project_dir):
            raise ValueError(f"Artifacts directory '{self._artifacts_dir}' "
                             f"must be within the project directory '{project_dir}'")

        if self._artifacts_dir.exists():
            shutil.rmtree(self._artifacts_dir)

    def on_scenario_run(self, event: ScenarioRunEvent) -> None:
        self._scenario_artifacts.clear()
        self._step_artifacts.clear()

    async def on_step_end(self, event: Union[StepPassedEvent, StepFailedEvent]) -> None:
        while len(self._step_artifacts) > 0:
            artifact = self._step_artifacts.popleft()
            event.step_result.attach(artifact)

    async def on_scenario_end(self,
                              event: Union[ScenarioPassedEvent, ScenarioFailedEvent]) -> None:
        while len(self._scenario_artifacts) > 0:
            artifact = self._scenario_artifacts.popleft()
            event.scenario_result.attach(artifact)

    async def on_scenario_reported(self, event: ScenarioReportedEvent) -> None:
        if not self._save_artifacts:
            return

        aggregated_result = event.aggregated_result
        for scenario_result in aggregated_result.scenario_results:
            scenario_artifacts_dir = self._get_scenario_artifacts_dir(scenario_result)

            for step_result in scenario_result.step_results:
                for artifact in step_result.artifacts:
                    artifact_path = self._save_artifact(artifact, scenario_artifacts_dir)
                    self._add_extra_details(step_result, artifact_path)

            for artifact in scenario_result.artifacts:
                artifact_path = self._save_artifact(artifact, scenario_artifacts_dir)
                self._add_extra_details(scenario_result, artifact_path)

    def _is_relative_to(self, path: Path, parent: Path) -> bool:
        try:
            path.relative_to(parent)
        except ValueError:
            return False
        else:
            return path != parent

    def _add_extra_details(self, result: Union[ScenarioResult, StepResult],
                           artifact_path: Path) -> None:
        if self._add_artifact_details:
            rel_path = artifact_path.relative_to(self._get_project_dir())
            result.add_extra_details(f"artifact '{rel_path}'")

    def _get_project_dir(self) -> Path:
        assert self._global_config is not None
        return self._global_config.project_dir.resolve()

    def _get_scenario_artifacts_dir(self, scenario_result: ScenarioResult) -> Path:
        scenario = scenario_result.scenario
        scenario_path = self._artifacts_dir / scenario.rel_path.with_suffix('')

        template_index = str(scenario.template_index or 0)
        started_at = str(scenario_result.started_at or 0).replace(".", "-")
        scenario_path /= f"{started_at}-{scenario.name}-{template_index}"

        return scenario_path

    def _save_artifact(self, artifact: Artifact, scenario_path: Path) -> Path:
        if not scenario_path.exists():
            scenario_path.mkdir(parents=True, exist_ok=True)

        if isinstance(artifact, MemoryArtifact):
            artifact_dest_path = (scenario_path / artifact.name).resolve()
            artifact_dest_path.write_bytes(artifact.data)
            return artifact_dest_path

        elif isinstance(artifact, FileArtifact):
            artifact_dest_path = (scenario_path / artifact.name).resolve()
            artifact_source_path = artifact.path
            if not artifact_source_path.is_absolute():
                artifact_source_path = (self._get_project_dir() / artifact_source_path).resolve()
            shutil.copy2(artifact_source_path, artifact_dest_path)
            return artifact_dest_path

        else:
            artifact_type = type(artifact).__name__
            rel_path = scenario_path.relative_to(self._get_project_dir())
            raise TypeError(f"Can't save artifact to '{rel_path}': unknown type '{artifact_type}'")


class Artifacted(PluginConfig):
    plugin = ArtifactedPlugin
    description = "Manages artifacts for step and scenario results"

    # Save artifacts to the file system
    save_artifacts: bool = False

    # Directory path for saving artifacts
    # Available if `save_artifacts` is True
    artifacts_dir: Path = Path(".vedro/artifacts/")

    # Add artifact details to scenario and steps extras
    # Available if `save_artifacts` is True
    add_artifact_details: bool = True



from abc import ABC, abstractmethod
from typing import Set

__all__ = ("TagMatcher",)


class TagMatcher(ABC):
    def __init__(self, expr: str) -> None:
        self._expr = expr

    @abstractmethod
    def match(self, tags: Set[str]) -> bool:
        pass

    @abstractmethod
    def validate(self, tag: str) -> bool:
        pass



import signal
from types import FrameType
from typing import Any, Dict, Optional, Tuple, Type, Union

from vedro.core import Dispatcher, Plugin, PluginConfig
from vedro.core.scenario_runner import Interrupted
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ScenarioReportedEvent,
    ScenarioRunEvent,
    ScenarioSkippedEvent,
    StartupEvent,
)

__all__ = ("Interrupter", "InterrupterPlugin", "InterrupterPluginTriggered",)


class InterrupterPluginTriggered(Interrupted):
    """
    Exception raised when the `InterrupterPlugin` conditions are met.

    This exception inherits from the `Interrupted` exception and is used
    specifically by the `InterrupterPlugin` when it is triggered. This could be
    due to the occurrence of specific signals that the plugin is set to handle, or
    when the limit of allowed scenario failures is reached during test execution.
    """
    pass


class InterrupterPlugin(Plugin):
    def __init__(self, config: Type["Interrupter"]) -> None:
        super().__init__(config)
        self._fail_fast: bool = False
        self._fail_after_count: Union[int, None] = None
        self._failed_count: int = 0
        self._signals: Tuple[signal.Signals, ...] = config.handle_signals
        self._orig_handlers: Dict[signal.Signals, Any] = {}

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(StartupEvent, self.on_startup) \
                  .listen(ScenarioRunEvent, self.on_scenario_execute) \
                  .listen(ScenarioSkippedEvent, self.on_scenario_execute) \
                  .listen(ScenarioReportedEvent, self.on_scenario_reported) \
                  .listen(CleanupEvent, self.on_cleanup)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        group = event.arg_parser.add_argument_group("Interrupter")
        exclusive_group = group.add_mutually_exclusive_group()

        # --ff will be removed in v2.0
        exclusive_group.add_argument("--fail-fast", "--ff", "-f",
                                     action="store_true",
                                     default=self._fail_fast,
                                     help="Stop after first failed scenario")
        exclusive_group.add_argument("--fail-after-count", type=int, default=None,
                                     help="Stop after N failed scenarios")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._fail_fast = event.args.fail_fast
        self._fail_after_count = event.args.fail_after_count

        if (self._fail_after_count is not None) and (self._fail_after_count < 1):
            raise ValueError("InterrupterPlugin: `fail_after_count` must be >= 1")

    def on_startup(self, event: StartupEvent) -> None:
        def handler(signum: int, frame: Optional[FrameType]) -> None:
            try:
                signame = signal.Signals(signum)
            except ValueError:
                signame = signum  # type: ignore
            raise InterrupterPluginTriggered(f"Stop after signal {signame!r} received")

        for sig in self._signals:
            self._orig_handlers[sig] = signal.getsignal(sig)
            signal.signal(sig, handler)

    def on_scenario_reported(self, event: ScenarioReportedEvent) -> None:
        if event.aggregated_result.is_failed():
            self._failed_count += 1

    def on_scenario_execute(self, event: Union[ScenarioRunEvent, ScenarioSkippedEvent]) -> None:
        if (not self._fail_fast) and (self._fail_after_count is None):
            return

        if self._fail_fast and self._failed_count >= 1:
            raise InterrupterPluginTriggered("Stop after first failed scenario")

        if self._fail_after_count and self._failed_count >= self._fail_after_count:
            raise InterrupterPluginTriggered(
                f"Stop after {self._fail_after_count} failed scenario")

    def on_cleanup(self, event: CleanupEvent) -> None:
        for sig, orig_handler in self._orig_handlers.items():
            signal.signal(sig, orig_handler)
        self._orig_handlers = {}


class Interrupter(PluginConfig):
    plugin = InterrupterPlugin
    description = "Halts test execution after N failed scenarios or on specified signals"

    # Raise Interrupted exception on these signals
    handle_signals: Tuple[signal.Signals, ...] = (signal.SIGTERM,)



from asyncio import CancelledError

import vedro.core as core
import vedro.plugins.artifacted as artifacted
import vedro.plugins.assert_rewriter as assert_rewriter
import vedro.plugins.deferrer as deferrer
import vedro.plugins.director as director
import vedro.plugins.dry_runner as dry_runner
import vedro.plugins.interrupter as interrupter
import vedro.plugins.last_failed as last_failed
import vedro.plugins.orderer as orderer
import vedro.plugins.repeater as repeater
import vedro.plugins.rerunner as rerunner
import vedro.plugins.seeder as seeder
import vedro.plugins.skipper as skipper
import vedro.plugins.slicer as slicer
import vedro.plugins.system_upgrade as system_upgrade
import vedro.plugins.tagger as tagger
import vedro.plugins.temp_keeper as temp_keeper
import vedro.plugins.terminator as terminator
from vedro.core import (
    Dispatcher,
    Factory,
    ModuleFileLoader,
    ModuleLoader,
    MonotonicScenarioRunner,
    MonotonicScenarioScheduler,
    MultiScenarioDiscoverer,
    ScenarioDiscoverer,
    ScenarioFileFinder,
    ScenarioFileLoader,
    ScenarioFinder,
    ScenarioLoader,
    ScenarioOrderer,
    ScenarioRunner,
    ScenarioScheduler,
    Singleton,
)
from vedro.core.scenario_finder.scenario_file_finder import (
    AnyFilter,
    DunderFilter,
    ExtFilter,
    HiddenFilter,
)
from vedro.core.scenario_orderer import StableScenarioOrderer

__all__ = ("Config",)


class Config(core.Config):

    # Validate each plugin's configuration, checking for unknown attributes to prevent errors
    validate_plugins_configs: bool = True

    class Registry(core.Config.Registry):
        Dispatcher = Singleton[Dispatcher](Dispatcher)

        ModuleLoader = Factory[ModuleLoader](ModuleFileLoader)

        ScenarioFinder = Factory[ScenarioFinder](lambda: ScenarioFileFinder(
            file_filter=AnyFilter([HiddenFilter(), DunderFilter(), ExtFilter(only=["py"])]),
            dir_filter=AnyFilter([HiddenFilter(), DunderFilter()])
        ))

        ScenarioLoader = Factory[ScenarioLoader](lambda: ScenarioFileLoader(
            module_loader=Config.Registry.ModuleLoader(),
        ))

        ScenarioOrderer = Factory[ScenarioOrderer](StableScenarioOrderer)

        ScenarioDiscoverer = Factory[ScenarioDiscoverer](lambda: MultiScenarioDiscoverer(
            finder=Config.Registry.ScenarioFinder(),
            loader=Config.Registry.ScenarioLoader(),
            orderer=Config.Registry.ScenarioOrderer(),
        ))

        ScenarioScheduler = Factory[ScenarioScheduler](MonotonicScenarioScheduler)

        # Note: The `ScenarioRunner` factory definition here should not be overridden
        # with a custom runner implementation because, starting from version 2.0,
        # custom runners will be deprecated and removed.
        ScenarioRunner = Factory[ScenarioRunner](lambda: MonotonicScenarioRunner(
            dispatcher=Config.Registry.Dispatcher(),
            interrupt_exceptions=(KeyboardInterrupt, SystemExit, CancelledError),
        ))

    class Plugins(core.Config.Plugins):
        class Director(director.Director):
            enabled = True

        class RichReporter(director.RichReporter):
            enabled = True

        class SilentReporter(director.SilentReporter):
            enabled = True

        class PyCharmReporter(director.PyCharmReporter):
            enabled = True

        class TempKeeper(temp_keeper.TempKeeper):
            enabled = True

        class Orderer(orderer.Orderer):
            enabled = True

        class LastFailed(last_failed.LastFailed):
            enabled = True

        class Deferrer(deferrer.Deferrer):
            enabled = True

        class Seeder(seeder.Seeder):
            enabled = True

        class Artifacted(artifacted.Artifacted):
            enabled = True

        class Skipper(skipper.Skipper):
            enabled = True

        class Slicer(slicer.Slicer):
            enabled = True

        class Tagger(tagger.Tagger):
            enabled = True

        class Repeater(repeater.Repeater):
            enabled = True

        class Rerunner(rerunner.Rerunner):
            enabled = True

        class AssertRewriter(assert_rewriter.AssertRewriter):
            enabled = True

        class DryRunner(dry_runner.DryRunner):
            enabled = True

        class Interrupter(interrupter.Interrupter):
            enabled = True

        class SystemUpgrade(system_upgrade.SystemUpgrade):
            enabled = True

        class Terminator(terminator.Terminator):
            enabled = True



import asyncio
from typing import Any, Callable, Coroutine, Type, Union

from vedro.core import ConfigType, Dispatcher, Plugin, PluginConfig, ScenarioScheduler
from vedro.core.scenario_runner import Interrupted
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ConfigLoadedEvent,
    ScenarioFailedEvent,
    ScenarioPassedEvent,
    ScenarioRunEvent,
    ScenarioSkippedEvent,
    StartupEvent,
)

from ._scheduler import RepeaterScenarioScheduler

__all__ = ("Repeater", "RepeaterPlugin", "RepeaterExecutionInterrupted",)


class RepeaterExecutionInterrupted(Interrupted):
    """
    Exception raised when the execution of scenario repetition is interrupted.

    This exception is used within the RepeaterPlugin to signal an early termination
    of the scenario repetition process. It is typically raised when the fail-fast
    condition is met, i.e., if a scenario fails and the --fail-fast-on-repeat option
    is enabled, indicating that further repetitions of the scenario should be stopped.
    """
    pass


SleepType = Callable[[float], Coroutine[Any, Any, None]]


class RepeaterPlugin(Plugin):
    def __init__(self, config: Type["Repeater"], *, sleep: SleepType = asyncio.sleep) -> None:
        super().__init__(config)
        self._scheduler_factory = config.scheduler_factory
        self._sleep = sleep
        self._repeats: int = 1
        self._repeats_delay: float = 0.0
        self._fail_fast: bool = False
        self._global_config: Union[ConfigType, None] = None
        self._scheduler: Union[ScenarioScheduler, None] = None
        self._repeat_scenario_id: Union[str, None] = None
        self._failed_count: int = 0

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ConfigLoadedEvent, self.on_config_loaded) \
                  .listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed) \
                  .listen(StartupEvent, self.on_startup) \
                  .listen(ScenarioRunEvent, self.on_scenario_execute) \
                  .listen(ScenarioSkippedEvent, self.on_scenario_execute) \
                  .listen(ScenarioPassedEvent, self.on_scenario_end) \
                  .listen(ScenarioFailedEvent, self.on_scenario_end) \
                  .listen(CleanupEvent, self.on_cleanup)

    def on_config_loaded(self, event: ConfigLoadedEvent) -> None:
        self._global_config = event.config

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        group = event.arg_parser.add_argument_group("Repeater")
        help_message = ("Number of times to repeat scenarios (default: 1). "
                        "The resulting scenario status will be failed if any repeat fail")
        group.add_argument("-N", "--repeats", type=int, default=self._repeats, help=help_message)
        group.add_argument("--repeats-delay", type=float, default=self._repeats_delay,
                           help="Delay in seconds between scenario repeats (default: 0.0s)")
        group.add_argument("--fail-fast-on-repeat", action="store_true", default=self._fail_fast,
                           help="Stop repeating scenarios after the first failure")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._repeats = event.args.repeats
        self._repeats_delay = event.args.repeats_delay
        self._fail_fast = event.args.fail_fast_on_repeat

        if self._repeats < 1:
            raise ValueError("--repeats must be >= 1")

        if self._repeats_delay < 0.0:
            raise ValueError("--repeats-delay must be >= 0.0")

        if (self._repeats_delay > 0.0) and (self._repeats <= 1):
            raise ValueError("--repeats-delay must be used with --repeats > 1")

        if self._fail_fast and (self._repeats <= 1):
            raise ValueError("--fail-fast-on-repeat must be used with --repeats > 1")

        if self._is_repeating_enabled():
            assert self._global_config is not None  # for type checking
            self._global_config.Registry.ScenarioScheduler.register(self._scheduler_factory, self)

    def on_startup(self, event: StartupEvent) -> None:
        self._scheduler = event.scheduler

    async def on_scenario_execute(self,
                                  event: Union[ScenarioRunEvent, ScenarioSkippedEvent]) -> None:
        if not self._is_repeating_enabled():
            return

        if self._fail_fast and (self._failed_count > 0):
            raise RepeaterExecutionInterrupted("Stop repeating scenarios after the first failure")

        scenario = event.scenario_result.scenario
        if (self._repeat_scenario_id == scenario.unique_id) and (self._repeats_delay > 0.0):
            await self._sleep(self._repeats_delay)

    async def on_scenario_end(self,
                              event: Union[ScenarioPassedEvent, ScenarioFailedEvent]) -> None:
        if not self._is_repeating_enabled():
            return
        assert isinstance(self._scheduler, RepeaterScenarioScheduler)  # for type checking

        scenario = event.scenario_result.scenario
        if scenario.unique_id != self._repeat_scenario_id:
            self._repeat_scenario_id = scenario.unique_id
            self._failed_count = 1 if event.scenario_result.is_failed() else 0
            for _ in range(self._repeats - 1):
                self._scheduler.schedule(scenario)
        else:
            if event.scenario_result.is_failed():
                self._failed_count += 1

    def on_cleanup(self, event: CleanupEvent) -> None:
        if not self._is_repeating_enabled():
            return

        if not event.report.interrupted:
            message = self._get_summary_message()
            event.report.add_summary(message)

    def _is_repeating_enabled(self) -> bool:
        return self._repeats > 1

    def _get_summary_message(self) -> str:
        message = f"repeated x{self._repeats}"
        if self._repeats_delay > 0.0:
            message += f" with delay {self._repeats_delay!r}s"
        return message


class Repeater(PluginConfig):
    plugin = RepeaterPlugin
    description = "Repeat scenarios a specified number of times"

    # Scheduler that will be used to create aggregated result for repeated scenarios
    scheduler_factory: Type[ScenarioScheduler] = RepeaterScenarioScheduler



import shutil
from typing import Type

from vedro.core import Dispatcher, Plugin, PluginConfig
from vedro.events import ArgParsedEvent, ConfigLoadedEvent

from ._temp_file_manager import TempFileManager

__all__ = ("TempKeeper", "TempKeeperPlugin", "create_tmp_dir", "create_tmp_file",)


_tmp_file_manager = TempFileManager()
create_tmp_dir = _tmp_file_manager.create_tmp_dir
create_tmp_file = _tmp_file_manager.create_tmp_file


class TempKeeperPlugin(Plugin):
    def __init__(self, config: Type["TempKeeper"], *,
                 tmp_file_manager: TempFileManager = _tmp_file_manager) -> None:
        super().__init__(config)
        self._tmp_file_manager = tmp_file_manager

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ConfigLoadedEvent, self.on_config_loaded) \
                  .listen(ArgParsedEvent, self.on_arg_parsed)

    def on_config_loaded(self, event: ConfigLoadedEvent) -> None:
        self._tmp_file_manager.set_project_dir(event.config.project_dir)

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        tmp_root = self._tmp_file_manager.get_tmp_root()
        if tmp_root.exists():
            shutil.rmtree(tmp_root)


class TempKeeper(PluginConfig):
    plugin = TempKeeperPlugin
    description = "Manages temporary directories and files"



import warnings
from pathlib import Path
from typing import Set, Type

from vedro import Config
from vedro.core import Dispatcher, MonotonicScenarioRunner, Plugin, PluginConfig
from vedro.events import (
    ArgParsedEvent,
    ArgParseEvent,
    CleanupEvent,
    ConfigLoadedEvent,
    StartupEvent,
)
from vedro.plugins.dry_runner import DryRunner

from .._cmd_arg_parser import CommandArgumentParser
from .._command import Command

__all__ = ("RunCommand",)


class RunCommand(Command):
    def __init__(self, config: Type[Config], arg_parser: CommandArgumentParser) -> None:
        """
        Initialize a new instance of the RunCommand class.

        :param config: Global configuration
        :param arg_parser: Argument parser for command-line options
        """
        super().__init__(config, arg_parser)

    async def _register_plugins(self, dispatcher: Dispatcher) -> None:
        """
        Register plugins with the dispatcher.

        :param dispatcher: Dispatcher to register plugins with
        """
        for _, section in self._config.Plugins.items():
            if not issubclass(section.plugin, Plugin) or (section.plugin is Plugin):
                raise TypeError(
                    f"Plugin {section.plugin} should be subclass of vedro.core.Plugin")

            if self._config.validate_plugins_configs:
                self._validate_plugin_config(section)

            if section.enabled:
                plugin = section.plugin(config=section)
                dispatcher.register(plugin)

    def _validate_plugin_config(self, plugin_config: Type[PluginConfig]) -> None:
        """
        Validate plugin's configuration.

        :param plugin_config: Configuration of the plugin.
        """
        unknown_attrs = self._get_attrs(plugin_config) - self._get_parent_attrs(plugin_config)
        if unknown_attrs:
            attrs = ", ".join(unknown_attrs)
            raise AttributeError(
                f"{plugin_config.__name__} configuration contains unknown attributes: {attrs}")

    def _get_attrs(self, cls: type) -> Set[str]:
        """
        Get the set of attributes for a class.

        :param cls: The class to get attributes for
        :return: The set of attribute names for the class
        """
        return set(vars(cls))

    def _get_parent_attrs(self, cls: type) -> Set[str]:
        """
        Recursively get attributes from parent classes.

        :param cls: The class to get parent attributes for
        :return: The set of attribute names for the parent classes
        """
        attrs = set()
        # `object` (the base for all classes) has no __bases__
        for base in cls.__bases__:
            attrs |= self._get_attrs(base)
            attrs |= self._get_parent_attrs(base)
        return attrs

    async def _parse_args(self, dispatcher: Dispatcher) -> None:
        """
        Parse command-line arguments and fire corresponding dispatcher events.

        :param dispatcher: The dispatcher to fire events
        """

        # Avoid unrecognized arguments error
        help_message = ("Specify the root directory of the project, used as a reference point for "
                        "relative paths and file operations. "
                        "Defaults to the directory from which the command is executed.")
        self._arg_parser.add_argument("--project-dir", type=Path, default=self._config.project_dir,
                                      help=help_message)

        # https://github.com/python/cpython/issues/95073
        self._arg_parser.remove_help_action()
        await dispatcher.fire(ArgParseEvent(self._arg_parser))
        self._arg_parser.add_help_action()

        args = self._arg_parser.parse_args()
        await dispatcher.fire(ArgParsedEvent(args))

    async def run(self) -> None:
        """
        Execute the command, including plugin registration, event dispatching,
        and scenario execution.
        """
        dispatcher = self._config.Registry.Dispatcher()
        await self._register_plugins(dispatcher)

        await dispatcher.fire(ConfigLoadedEvent(self._config.path, self._config))

        await self._parse_args(dispatcher)

        start_dir = Path("scenarios")
        discoverer = self._config.Registry.ScenarioDiscoverer()
        try:
            scenarios = await discoverer.discover(start_dir)
        except SystemExit as e:
            raise Exception(f"SystemExit({e.code}) ⬆")

        scheduler = self._config.Registry.ScenarioScheduler(scenarios)
        await dispatcher.fire(StartupEvent(scheduler))

        runner = self._config.Registry.ScenarioRunner()
        if not isinstance(runner, (MonotonicScenarioRunner, DryRunner)):
            warnings.warn("Deprecated: custom runners will be removed in v2.0", DeprecationWarning)
        report = await runner.run(scheduler)

        await dispatcher.fire(CleanupEvent(report))



from types import TracebackType
from typing import Type

__all__ = ("ExcInfo",)


class ExcInfo:
    """
    Represent exception information.

    This class encapsulates the details of an exception, including its type, the exception
    instance itself, and the traceback associated with the exception. It provides a structured
    way to store and access exception information.
    """

    def __init__(self,
                 type_: Type[BaseException],
                 value: BaseException,
                 traceback: TracebackType) -> None:
        """
        Initialize an instance of ExcInfo with exception details.

        :param type_: The type of the exception.
        :param value: The exception instance.
        :param traceback: The traceback object associated with the exception.
        """
        self.type = type_
        self.value = value
        self.traceback = traceback

    def __repr__(self) -> str:
        """
        Return a string representation of the ExcInfo instance.

        :return: A string representation of the ExcInfo instance.
        """
        return f"{self.__class__.__name__}({self.type!r}, {self.value!r}, {self.traceback!r})"



from enum import Enum
from typing import Any, Callable, Type, Union

from vedro.core import Dispatcher, Plugin, PluginConfig, VirtualScenario
from vedro.events import ArgParsedEvent, ArgParseEvent, StartupEvent

from ._tag_matcher import TagMatcher
from .logic_tag_matcher import LogicTagMatcher

__all__ = ("Tagger", "TaggerPlugin",)


class TaggerPlugin(Plugin):
    def __init__(self, config: Type["Tagger"], *,
                 tag_matcher_factory: Callable[[str], TagMatcher] = LogicTagMatcher) -> None:
        super().__init__(config)
        self._matcher_factory = tag_matcher_factory
        self._matcher: Union[TagMatcher, None] = None
        self._tags_expr: Union[str, None] = None

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
            .listen(ArgParsedEvent, self.on_arg_parsed) \
            .listen(StartupEvent, self.on_startup)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        event.arg_parser.add_argument("-t", "--tags",
                                      help="Specify tags to selectively run scenarios")

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._tags_expr = event.args.tags

    def _get_tags(self, scenario: VirtualScenario, validate: Callable[[str], bool]) -> Any:
        orig_tags = getattr(scenario._orig_scenario, "tags", ())
        if not isinstance(orig_tags, (list, tuple, set)):
            raise TypeError(f"Scenario '{scenario.unique_id}' tags must be a list, tuple or set, "
                            f"got {type(orig_tags)}")
        tags = []
        for tag in orig_tags:
            if isinstance(tag, Enum):
                tag = tag.value
            try:
                validate(tag)
            except Exception as e:
                raise ValueError(f"Scenario '{scenario.unique_id}' tag '{tag}' is not valid ({e})")
            else:
                tags.append(tag)
        return tags

    async def on_startup(self, event: StartupEvent) -> None:
        self._matcher = self._matcher_factory(str(self._tags_expr))

        async for scenario in event.scheduler:
            tags = self._get_tags(scenario, self._matcher.validate)
            if self._tags_expr and not self._matcher.match(set(tags)):
                event.scheduler.ignore(scenario)


class Tagger(PluginConfig):
    plugin = TaggerPlugin
    description = "Allows scenarios to be selectively run based on user-defined tags"



from abc import ABC, abstractmethod
from typing import Set

__all__ = ("And", "Or", "Not", "Operand", "Operator", "Expr",)


class Expr(ABC):
    @abstractmethod
    def __call__(self, tags: Set[str]) -> bool:
        pass

    @abstractmethod
    def __repr__(self) -> str:
        pass


class Operator(Expr, ABC):
    pass


class Operand(Expr):
    def __init__(self, tag: str) -> None:
        self._tag = tag

    def __call__(self, tags: Set[str]) -> bool:
        return self._tag in tags

    def __repr__(self) -> str:
        return f"Tag({self._tag})"


class And(Operator):
    def __init__(self, left: Operand, right: Operand) -> None:
        self._left = left
        self._right = right

    def __call__(self, tags: Set[str]) -> bool:
        return self._left(tags) and self._right(tags)

    def __repr__(self) -> str:
        return f"And({self._left}, {self._right})"


class Or(Operator):
    def __init__(self, left: Operand, right: Operand) -> None:
        self._left = left
        self._right = right

    def __call__(self, tags: Set[str]) -> bool:
        return self._left(tags) or self._right(tags)

    def __repr__(self) -> str:
        return f"Or({self._left}, {self._right})"


class Not(Operator):
    def __init__(self, operand: Operand) -> None:
        self._operand = operand

    def __call__(self, tags: Set[str]) -> bool:
        return not self._operand(tags)

    def __repr__(self) -> str:
        return f"Not({self._operand})"



import string
import sys
import uuid
from hashlib import blake2b
from typing import Dict, Type, Union

from vedro.core import Dispatcher, Plugin, PluginConfig
from vedro.events import ArgParsedEvent, ArgParseEvent, CleanupEvent, ScenarioRunEvent

from ._random import RandomGenerator, StandardRandomGenerator

__all__ = ("Seeder", "SeederPlugin",)

_random = StandardRandomGenerator()


class SeederPlugin(Plugin):
    def __init__(self, config: Type["Seeder"], *, random: RandomGenerator = _random) -> None:
        super().__init__(config)
        self._random = random
        self._use_fixed_seed = config.use_fixed_seed
        self._show_seeds = config.show_seeds
        self._initial_seed: Union[str, None] = None
        self._history: Dict[str, int] = {}

    def subscribe(self, dispatcher: Dispatcher) -> None:
        dispatcher.listen(ArgParseEvent, self.on_arg_parse) \
                  .listen(ArgParsedEvent, self.on_arg_parsed, priority=-sys.maxsize - 1) \
                  .listen(ScenarioRunEvent, self.on_scenario_run) \
                  .listen(CleanupEvent, self.on_cleanup)

    def on_arg_parse(self, event: ArgParseEvent) -> None:
        event.arg_parser.add_argument("--seed", type=str, nargs="?", help="Set seed")

        help_msg = "Use the same seed when a scenario is run multiple times in the same execution"
        event.arg_parser.add_argument("--fixed-seed",
                                      action="store_true",
                                      default=self._use_fixed_seed,
                                      help=help_msg)

        help_msg = "Show concrete seeds for each scenario run"
        event.arg_parser.add_argument("--show-seeds",
                                      action="store_true",
                                      default=self._show_seeds,
                                      help=help_msg)

    def on_arg_parsed(self, event: ArgParsedEvent) -> None:
        self._initial_seed = event.args.seed or self._gen_initial_seed()
        self._use_fixed_seed = event.args.fixed_seed
        self._show_seeds = event.args.show_seeds

        self._random.set_seed(self._initial_seed)

    def on_scenario_run(self, event: ScenarioRunEvent) -> None:
        assert self._initial_seed is not None  # for type checker

        unique_id = event.scenario_result.scenario.unique_id
        self._history[unique_id] = self._history.get(unique_id, 0) + 1

        index = 1 if self._use_fixed_seed else self._history[unique_id]
        seed = self._create_seed(self._initial_seed, unique_id, index)
        self._random.set_seed(seed)

        if self._show_seeds:
            seed_repr = self._get_seed_repr(seed)
            event.scenario_result.add_extra_details(f"seed: {seed_repr[:8]}..{seed_repr[-8:]}")

    def on_cleanup(self, event: CleanupEvent) -> None:
        if (event.report.passed + event.report.failed) > 0:
            assert self._initial_seed is not None  # for type checker
            event.report.add_summary("--seed " + self._get_seed_repr(self._initial_seed))

    def _get_seed_repr(self, seed: str) -> str:
        alphabet = set(string.ascii_letters + string.digits + "-_")
        for char in seed:
            if char not in alphabet:
                return f"{seed!r}"
        return f"{seed}"

    def _gen_initial_seed(self) -> str:
        return str(uuid.uuid4())

    def _create_seed(self, initial_seed: str, scenario_id: str, index: int) -> str:
        seed = f"{initial_seed}//{scenario_id}//{index}"
        return blake2b(seed.encode()).hexdigest()


class Seeder(PluginConfig):
    plugin = SeederPlugin
    description = "Sets seeds for deterministic random behavior in scenarios"

    # Use the same seed when a scenario is run multiple times in the same execution
    use_fixed_seed: bool = False

    # Show concrete seeds for each scenario run
    show_seeds: bool = False
